#include <stdlib.h>
#include <stdint.h>

#include "../../include/inc_libc64_mini.h"
#include "hqx_dsp.h"

static int dsp_initialized = 0;
static dsp_component_id_t compid_hqx_dsp;
static dsp_mem_region_t scratchmem, l1mem, srcmem, dstmem;
static unsigned int inv_ptr, inv_size;

void hq2x_16_Init(uint32_t *hqTable);
void hq3x_16_Init(uint32_t *hqTable);

uint32_t arm_hq2x_8_16_partial(const uint8_t *src, uint16_t *dst, const uint32_t *palette, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height, volatile uint32_t *countptr, volatile uint32_t *stopptr);
uint32_t arm_hq2x_16_16_partial(const uint16_t *src, uint16_t *dst, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height, volatile uint32_t *countptr, volatile uint32_t *stopptr);

uint32_t arm_hq3x_8_16_partial(const uint8_t *src, uint16_t *dst, const uint32_t *palette, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height, volatile uint32_t *countptr, volatile uint32_t *stopptr);
uint32_t arm_hq3x_16_16_partial(const uint16_t *src, uint16_t *dst, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height, volatile uint32_t *countptr, volatile uint32_t *stopptr);

void neon_normal1x_8_8(const uint8_t *src, uint8_t *dst, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height);
void neon_normal1x_16_16(const uint16_t *src, uint16_t *dst, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height);


void dsp_hq2x_8_16(const uint8_t *src, uint16_t *dst, const uint32_t *palette, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height)
{
    dsp_msg_t msg;
    dsp_msg_t reply;
    int ret;
    scaler_params_t *params;
    uint8_t *dspsrc;
    uint16_t *dspdst;
    uint16_t *dsppalette;
    unsigned int dspsrcstride, dspdststride;

    volatile uint32_t *countptr;
    volatile uint32_t *stopptr;
    uint32_t count1;

    dspsrc = (uint8_t *)(srcmem.virt_addr);
    dspdst = (uint16_t *)(dstmem.virt_addr);
    dsppalette = (uint16_t *)(l1mem.virt_addr + 256*4 + 256*4);

    dspsrcstride = width * sizeof(uint8_t);
    dspdststride = 2 * width * sizeof(uint16_t);


    DSP_MSG_INIT(&msg, compid_hqx_dsp, HQX_DSP_HQ2X_8_16,
                 dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4 + 256*4 + 256*2),           // argument #1
                 0            // argument #2
                );

    params = (scaler_params_t *)(l1mem.virt_addr + 256*4 + 256*4 + 256*2);
    params->src = (void *) dsp_physgpp_to_physdsp(srcmem.phys_addr);
    params->dst = (void *) dsp_physgpp_to_physdsp(dstmem.phys_addr);
    params->hqtable = (void *) dsp_physgpp_to_physdsp(l1mem.phys_addr);
    params->palette = (void *) dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4 + 256*4);
    params->width = width;
    params->srcstride = dspsrcstride;
    params->dststride = dspdststride;
    params->height = height;

    // copy src to dspsrc
    neon_normal1x_8_8(src, dspsrc, width, srcstride, dspsrcstride, height);

    // copy palette to dsppalette
    for (count1 = 0; count1 < 256; count1++)
    {
        dsppalette[count1] = palette[count1];
    }

    countptr = (uint32_t *)(scratchmem.virt_addr + 4096);
    stopptr = (uint32_t *)(((uintptr_t)countptr) + 64);

    *countptr = 0;
    *stopptr = 0;

    ret = dsp_rpc_send(&msg);

    if (0 != ret)
    {
        return;
    }

    // invalidate cache for data from previous run
    if (inv_size != 0)
    {
        dsp_cache_inv(inv_ptr, inv_size);
        inv_size = 0;
    }

    count1 = arm_hq2x_8_16_partial(src, dst, palette, width, srcstride, dststride, height, countptr, stopptr);
    if (count1 < height)
    {
        // remember data for later cache invalidation
        inv_ptr = dstmem.phys_addr + count1 * 2 * dspdststride;
        inv_size = dspdststride * 2 * (height - count1);


        // copy dspdst to dst
        neon_normal1x_16_16(dspdst + count1 * dspdststride, dst + count1 * dststride, 2 * width, dspdststride, dststride, 2 * (height - count1));
    }

    dsp_rpc_recv(&reply);
}

void dsp_hq2x_16_16(const uint16_t *src, uint16_t *dst, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height)
{
    dsp_msg_t msg;
    dsp_msg_t reply;
    int ret;
    scaler_params_t *params;
    uint16_t *dspsrc;
    uint16_t *dspdst;
    unsigned int dspsrcstride, dspdststride;

    volatile uint32_t *countptr;
    volatile uint32_t *stopptr;
    uint32_t count1;

    dspsrc = (uint16_t *)(srcmem.virt_addr);
    dspdst = (uint16_t *)(dstmem.virt_addr);

    dspsrcstride = width * sizeof(uint16_t);
    dspdststride = 2 * width * sizeof(uint16_t);


    DSP_MSG_INIT(&msg, compid_hqx_dsp, HQX_DSP_HQ2X_16_16,
                 dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4 + 256*4 + 256*2),           // argument #1
                 0            // argument #2
                );

    params = (scaler_params_t *)(l1mem.virt_addr + 256*4 + 256*4 + 256*2);
    params->src = (void *) dsp_physgpp_to_physdsp(srcmem.phys_addr);
    params->dst = (void *) dsp_physgpp_to_physdsp(dstmem.phys_addr);
    params->hqtable = (void *) dsp_physgpp_to_physdsp(l1mem.phys_addr);
    params->palette = NULL;
    params->width = width;
    params->srcstride = dspsrcstride;
    params->dststride = dspdststride;
    params->height = height;

    // copy src to dspsrc
    neon_normal1x_16_16(src, dspsrc, width, srcstride, dspsrcstride, height);

    countptr = (uint32_t *)(scratchmem.virt_addr + 4096);
    stopptr = (uint32_t *)(((uintptr_t)countptr) + 64);

    *countptr = 0;
    *stopptr = 0;

    ret = dsp_rpc_send(&msg);

    if (0 != ret)
    {
        return;
    }

    // invalidate cache for data from previous run
    if (inv_size != 0)
    {
        dsp_cache_inv(inv_ptr, inv_size);
        inv_size = 0;
    }

    count1 = arm_hq2x_16_16_partial(src, dst, width, srcstride, dststride, height, countptr, stopptr);
    if (count1 < height)
    {
        // remember data for later cache invalidation
        inv_ptr = dstmem.phys_addr + count1 * 2 * dspdststride;
        inv_size = dspdststride * 2 * (height - count1);

        // copy dspdst to dst
        neon_normal1x_16_16(dspdst + count1 * dspdststride, dst + count1 * dststride, 2 * width, dspdststride, dststride, 2 * (height - count1));
    }

    dsp_rpc_recv(&reply);
}

void dsp_hq3x_8_16(const uint8_t *src, uint16_t *dst, const uint32_t *palette, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height)
{
    dsp_msg_t msg;
    dsp_msg_t reply;
    int ret;
    scaler_params_t *params;
    uint8_t *dspsrc;
    uint16_t *dspdst;
    uint16_t *dsppalette;
    unsigned int dspsrcstride, dspdststride;

    volatile uint32_t *countptr;
    volatile uint32_t *stopptr;
    uint32_t count1;

    dspsrc = (uint8_t *)(srcmem.virt_addr);
    dspdst = (uint16_t *)(dstmem.virt_addr);
    dsppalette = (uint16_t *)(l1mem.virt_addr + 256*4 + 256*4);

    dspsrcstride = width * sizeof(uint8_t);
    dspdststride = 3 * width * sizeof(uint16_t);


    DSP_MSG_INIT(&msg, compid_hqx_dsp, HQX_DSP_HQ3X_8_16,
                 dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4 + 256*4 + 256*2),           // argument #1
                 0            // argument #2
                );

    params = (scaler_params_t *)(l1mem.virt_addr + 256*4 + 256*4 + 256*2);
    params->src = (void *) dsp_physgpp_to_physdsp(srcmem.phys_addr);
    params->dst = (void *) dsp_physgpp_to_physdsp(dstmem.phys_addr);
    params->hqtable = (void *) dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4);
    params->palette = (void *) dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4 + 256*4);
    params->width = width;
    params->srcstride = dspsrcstride;
    params->dststride = dspdststride;
    params->height = height;

    // copy src to dspsrc
    neon_normal1x_8_8(src, dspsrc, width, srcstride, dspsrcstride, height);

    // copy palette to dsppalette
    for (count1 = 0; count1 < 256; count1++)
    {
        dsppalette[count1] = palette[count1];
    }

    countptr = (uint32_t *)(scratchmem.virt_addr + 4096);
    stopptr = (uint32_t *)(((uintptr_t)countptr) + 64);

    *countptr = 0;
    *stopptr = 0;

    ret = dsp_rpc_send(&msg);

    if (0 != ret)
    {
        return;
    }

    // invalidate cache for data from previous run
    if (inv_size != 0)
    {
        dsp_cache_inv(inv_ptr, inv_size);
        inv_size = 0;
    }

    count1 = arm_hq3x_8_16_partial(src, dst, palette, width, srcstride, dststride, height, countptr, stopptr);
    if (count1 < height)
    {
        // remember data for later cache invalidation
        inv_ptr = dstmem.phys_addr + count1 * 3 * dspdststride;
        inv_size = dspdststride * 3 * (height - count1);

        // copy dspdst to dst
        neon_normal1x_16_16(dspdst + count1 * 3 * dspdststride / sizeof(uint16_t), dst + count1 * 3 * dststride / sizeof(uint16_t), 3 * width, dspdststride, dststride, 3 * (height - count1));
    }

    dsp_rpc_recv(&reply);
}

void dsp_hq3x_16_16(const uint16_t *src, uint16_t *dst, unsigned int width, unsigned int srcstride, unsigned int dststride, unsigned int height)
{
    dsp_msg_t msg;
    dsp_msg_t reply;
    int ret;
    scaler_params_t *params;
    uint16_t *dspsrc;
    uint16_t *dspdst;
    unsigned int dspsrcstride, dspdststride;

    volatile uint32_t *countptr;
    volatile uint32_t *stopptr;
    uint32_t count1;

    dspsrc = (uint16_t *)(srcmem.virt_addr);
    dspdst = (uint16_t *)(dstmem.virt_addr);

    dspsrcstride = width * sizeof(uint16_t);
    dspdststride = 3 * width * sizeof(uint16_t);


    DSP_MSG_INIT(&msg, compid_hqx_dsp, HQX_DSP_HQ3X_16_16,
                 dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4 + 256*4 + 256*2),           // argument #1
                 0            // argument #2
                );

    params = (scaler_params_t *)(l1mem.virt_addr + 256*4 + 256*4 + 256*2);
    params->src = (void *) dsp_physgpp_to_physdsp(srcmem.phys_addr);
    params->dst = (void *) dsp_physgpp_to_physdsp(dstmem.phys_addr);
    params->hqtable = (void *) dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4);
    params->palette = NULL;
    params->width = width;
    params->srcstride = dspsrcstride;
    params->dststride = dspdststride;
    params->height = height;

    // copy src to dspsrc
    neon_normal1x_16_16(src, dspsrc, width, srcstride, dspsrcstride, height);

    countptr = (uint32_t *)(scratchmem.virt_addr + 4096);
    stopptr = (uint32_t *)(((uintptr_t)countptr) + 64);

    *countptr = 0;
    *stopptr = 0;

    ret = dsp_rpc_send(&msg);

    if (0 != ret)
    {
        return;
    }

    // invalidate cache for data from previous run
    if (inv_size != 0)
    {
        dsp_cache_inv(inv_ptr, inv_size);
        inv_size = 0;
    }

    count1 = arm_hq3x_16_16_partial(src, dst, width, srcstride, dststride, height, countptr, stopptr);
    if (count1 < height)
    {
        // remember data for later cache invalidation
        inv_ptr = dstmem.phys_addr + count1 * 3 * dspdststride;
        inv_size = dspdststride * 3 * (height - count1);

        // copy dspdst to dst
        neon_normal1x_16_16(dspdst + count1 * 3 * dspdststride / sizeof(uint16_t), dst + count1 * 3 * dststride / sizeof(uint16_t), 3 * width, dspdststride, dststride, 3 * (height - count1));
    }

    dsp_rpc_recv(&reply);
}

void dsp_hqx_deinitialize(void)
{
    if (!dsp_initialized) return;

    dsp_initialized = 0;

    dsp_shm_free(dstmem);
    dsp_shm_free(srcmem);
    dsp_l1sram_free(l1mem);

    dsp_close();
}

int dsp_hqx_initialize(unsigned int maxwidth, unsigned int maxheight)
{
    int ret;

    if (dsp_initialized) return 1;

    if ((maxwidth > 320) || (maxwidth < 32) || (maxheight < 5)) return 2;

    system("sudo /usr/pandora/scripts/op_dsp_c64.sh");

    ret = dsp_open();
    if (0 != ret)
    {
        return 3;
    }

    if (0 != dsp_component_load(NULL, COMPONENT_NAME_HQX_DSP, &compid_hqx_dsp))
    {
        dsp_close();
        return 4;
    }


    maxwidth = (maxwidth + 7) & ~7;

    scratchmem = dsp_l1sram_scratchbuffer_get();
    if (0 == scratchmem.size)
    {
        dsp_close();
        return 5;
    }

    l1mem = dsp_l1sram_alloc(256*4 + 256*4 + 256*2 + sizeof(scaler_params_t));
    if (0 == l1mem.size)
    {
        dsp_close();
        return 6;
    }

    srcmem = dsp_shm_alloc(DSP_CACHE_W, maxwidth*maxheight*2);
    if (0 == srcmem.size)
    {
        dsp_l1sram_free(l1mem);
        dsp_close();
        return 7;
    }

    dstmem = dsp_shm_alloc(DSP_CACHE_R, 3*maxwidth*3*maxheight*2);
    if (0 == dstmem.size)
    {
        dsp_shm_free(srcmem);
        dsp_l1sram_free(l1mem);
        dsp_close();
        return 8;
    }


    hq2x_16_Init((uint32_t *) l1mem.virt_addr);
    hq3x_16_Init(((uint32_t *) l1mem.virt_addr) + 256);

    // init scalers on dsp
    {
        dsp_msg_t msg;
        dsp_msg_t reply;
        scaler_params_t *params;

        DSP_MSG_INIT(&msg, compid_hqx_dsp, HQX_DSP_INIT_HQX,
                     dsp_physgpp_to_physdsp(l1mem.phys_addr + 256*4 + 256*4 + 256*2),           // argument #1
                     0            // argument #2
                    );

        params = (scaler_params_t *)(l1mem.virt_addr + 256*4 + 256*4 + 256*2);
        params->src = (void *) dsp_physgpp_to_physdsp(srcmem.phys_addr);
        params->dst = (void *) dsp_physgpp_to_physdsp(dstmem.phys_addr);
        params->hqtable = NULL;
        params->palette = NULL;
        params->width = maxwidth;
        params->srcstride = maxwidth*sizeof(uint16_t);
        params->dststride = 3*maxwidth*sizeof(uint16_t);
        params->height = maxheight;

        ret = dsp_rpc_send(&msg);

        if (0 != ret)
        {
            dsp_shm_free(dstmem);
            dsp_shm_free(srcmem);
            dsp_l1sram_free(l1mem);

            dsp_close();
            return 9;
        }

        // invalidate cache for dstmem
        dsp_cache_inv(dstmem.phys_addr, 3*maxwidth*sizeof(uint16_t) * 3 * maxheight);
        inv_size = 0;

        ret = dsp_rpc_recv(&reply);

        if (0 != ret)
        {
            dsp_shm_free(dstmem);
            dsp_shm_free(srcmem);
            dsp_l1sram_free(l1mem);

            dsp_close();
            return 10;
        }
    }

    dsp_initialized = 1;

    atexit(dsp_hqx_deinitialize);

    return 0;
}

